from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class TopologicalTorsionFingerprint(BaseFingerprintTransformer):
    """
    Topological Torsion fingerprint.

    The implementation uses RDKit. This is a hashed fingerprint, where
    the hashed fragments are computed based on topological torsions [1]_.

    A topological torsion is defined as a linear sequence of consecutively bonded heavy (non-hydrogen):
    (atom 1 type)-(atom 2 type)-(atom 3 type)-(atom 4 type)

    Atom type takes into consideration:

    - atomic number
    - number of pi electrons
    - degree (number of bonds)

    This example of 4 atom path is the canonical version of topological torsion.
    The number of atoms can be adjusted (using ``torsion_atom_count`` parameter).

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    torsion_atom_count : int, default=4
        The number of atoms to be included in the torsion.

    use_pharmacophoric_invariants: bool, default=False
        Whether to use pharmacophoric invariants (atom types) instead of default ones.
        They are the same as in the FCFP fingerprint: Donor, Acceptor, Aromatic,
        Halogen, Basic, Acidic.

    include_chirality : bool, default=False
        Whether to include chirality information when computing atom types.

    count_simulation : bool, default=True
        Whether to use count simulation for approximating feature counts [2]_.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`AtomPairFingerprint` : Related fingerprint, but uses 2 atoms and the distance between them.

    References
    ----------
    .. [1] `Ramaswamy Nilakantan, Norman Bauman, J. Scott Dixon, R. Venkataraghavan
        "Topological torsion: a new molecular descriptor for SAR applications. Comparison with other descriptors"
        J. Chem. Inf. Comput. Sci. 1987, 27, 82-85
        <https://pubs.acs.org/doi/10.1021/ci00054a008>`_

    .. [2] `Gregory Landrum
        "Simulating count fingerprints"
        RDKit blog 2021
        <https://greglandrum.github.io/rdkit-blog/posts/2021-07-06-simulating-counts.html>`_

    Examples
    --------
    >>> from skfp.fingerprints import TopologicalTorsionFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = TopologicalTorsionFingerprint()
    >>> fp
    TopologicalTorsionFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)

    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "torsion_atom_count": [Interval(Integral, 2, None, closed="left")],
        "use_pharmacophoric_invariants": ["boolean"],
        "include_chirality": ["boolean"],
        "count_simulation": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        torsion_atom_count: int = 4,
        use_pharmacophoric_invariants: bool = False,
        include_chirality: bool = False,
        count_simulation: bool = True,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.torsion_atom_count = torsion_atom_count
        self.use_pharmacophoric_invariants = use_pharmacophoric_invariants
        self.include_chirality = include_chirality
        self.count_simulation = count_simulation

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute Topological Torsion fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.
        """
        return super().transform(X, copy)

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdFingerprintGenerator import (
            GetMorganFeatureAtomInvGen,
            GetTopologicalTorsionGenerator,
        )

        X = ensure_mols(X)

        if self.use_pharmacophoric_invariants:
            inv_gen = GetMorganFeatureAtomInvGen()
        else:
            inv_gen = None

        gen = GetTopologicalTorsionGenerator(
            fpSize=self.fp_size,
            torsionAtomCount=self.torsion_atom_count,
            atomInvariantsGenerator=inv_gen,
            includeChirality=self.include_chirality,
            countSimulation=self.count_simulation,
        )

        if self.count:
            X = [gen.GetCountFingerprintAsNumPy(mol) for mol in X]
        else:
            X = [gen.GetFingerprintAsNumPy(mol) for mol in X]

        dtype = np.uint32 if self.count else np.uint8
        if self.sparse:
            return csr_array(X, dtype=dtype)
        else:
            return np.array(X, dtype=dtype)



from collections.abc import Sequence
from numbers import Integral, Real
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError, StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class ERGFingerprint(BaseFingerprintTransformer):
    """
    Extended-Reduced Graph (ERG) fingerprint.

    The implementation uses RDKit. This fingerprint is descriptor-based, but has
    variable length, depending on distance parameters. Originally described in [1]_.

    This fingerprint can be seen as a hybrid of reduced graphs [2]_ and binding
    property (BP) pairs [3]_. It uses fuzzy incrementation instead of counts.

    First, input molecule is converted into a "reduced graph", condensing
    pharmacophorically relevant information. On that graph, property points (PP) are
    identified. From those nodes, triplets (PP 1, topological distance, PP 2) are
    formed, using the shortest path distances on the reduced graph. They are not
    hashed, in contrast to Atom Pair approach, but rather all such triplets with
    distances between ``min_distance`` and ``max_distance`` are considered. For each
    triplet occurrence, its bit is incremented by 1 (resulting in triplet counts),
    and additionally, the fields for closest distances, i.e. ``dist-1`` and ``dist+1``,
    are also incremented by ``fuzz_increment`` (hence the fuzziness).

    Six features are used for both computing the reduced graph and identifying
    property points, using SMARTS patterns definitions from [4]_:
    - hydrogen bond donors
    - hydrogen bond acceptors
    - aromatic ring systems
    - hydrophobic (aliphatic) ring systems
    - positive formal atom charges
    - negative formal atom charges

    Because we always use 6 features, the resulting fingerprints have length
    ``21 * (max_distance - min_distance + 1)``.

    Note that RDKit does not implement two features from the original paper,
    flip-flop flags (for donors and acceptors), and collapsing highly fused rings.

    Parameters
    ----------
    fuzz_increment : float, default=0.3
        How much to increment triplets occurrences and their closest neighboring
        fields by when they are detected. Controls how much weight is put on the
        similarity of pharmacophoric patterns. Default value of 0.3 was optimized
        for scaffold-hopping applications, but lower values around 0.1-0.2 can be
        considered for more "crisp" similarity.

    min_path : int, default=1
        Minimal shortest path length to consider for calculating triplets on the
        reduced graph.

    max_path : int, default=15
        Maximal shortest path length to consider for calculating triplets on the
        reduced graph.

    variant : {"fuzzy", "bit", "count"}, default="fuzzy"
        Fingerprint variant. Default "fuzzy" follows the original paper and results
        in floating point results. "bit" and "count" result in integer vectors for
        "crisp" calculation on the original graph, and they use zero fuzziness,
        ignoring the ``fuzz_incremement`` parameter.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to
        ``21 * (max_path - min_path)``, 315 for default parameters.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`TopologicalTorsionFingerprint` : Related fingerprint, but uses 4-atom paths.

    References
    ----------
    .. [1] `Nikolaus Stiefl, Ian A. Watson, Knut Baumann, and Andrea Zaliani
        "ErG: 2D Pharmacophore Descriptions for Scaffold Hopping"
        J. Chem. Inf. Model. 2006, 46, 1, 208–220
        <https://pubs.acs.org/doi/abs/10.1021/ci050457y>`_

    .. [2] `Valerie J. Gillet, Peter Willett, and John Bradshaw
        "Similarity Searching Using Reduced Graphs"
        J. Chem. Inf. Comput. Sci. 2003, 43, 2, 338–345
        <https://pubs.acs.org/doi/abs/10.1021/ci025592e>`_

    .. [3] `Simon K. Kearsley, Susan Sallamack, Eugene M. Fluder, Joseph D. Andose,
        Ralph T. Mosley, and Robert P. Sheridan
        "Chemical Similarity Using Physiochemical Property Descriptors"
        J. Chem. Inf. Comput. Sci. 1996, 36, 1, 118–127
        <https://pubs.acs.org/doi/abs/10.1021/ci950274j>`_

    .. [4] `Alberto Gobbi and Dieter Poppinger
        "Genetic optimization of combinatorial libraries"
        Biotechnol Bioeng. 1998 Winter;61(1):47-54.
        <https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/abs/10.1002/(SICI)1097-0290(199824)61:1%3C47::AID-BIT9%3E3.0.CO%3B2-Z>`_

    Examples
    --------
    >>> from skfp.fingerprints import ERGFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = ERGFingerprint()
    >>> fp
    ERGFingerprint()

    >>> fp.transform(smiles)
    array([[0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fuzz_increment": [Interval(Real, 0.0, None, closed="left")],
        "min_path": [Interval(Integral, 1, None, closed="left")],
        "max_path": [Interval(Integral, 1, None, closed="left")],
        "variant": [StrOptions({"fuzzy", "bit", "count"})],
    }

    def __init__(
        self,
        fuzz_increment: float = 0.3,
        min_path: int = 1,
        max_path: int = 15,
        variant: str = "fuzzy",
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=int(max_path * 21),
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fuzz_increment = fuzz_increment
        self.min_path = min_path
        self.max_path = max_path
        self.variant = variant

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.max_path <= self.min_path:
            raise InvalidParameterError(
                f"The max_path parameter of {self.__class__.__name__} must be "
                f"greater than min_path, got: "
                f"min_path={self.min_path}, max_path={self.max_path}"
            )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdReducedGraphs import GetErGFingerprint

        X = ensure_mols(X)

        fuzz = self.fuzz_increment if self.variant == "fuzzy" else 0

        X = np.array(
            [
                GetErGFingerprint(
                    mol,
                    fuzzIncrement=fuzz,
                    minPath=self.min_path,
                    maxPath=self.max_path,
                )
                for mol in X
            ]
        )

        if self.variant == "bit":
            X = X > 0
            dtype = np.uint8
        elif self.variant == "count":
            X = np.round(X)
            dtype = np.uint32  # type: ignore
        else:  # "fuzzy"
            dtype = float  # type: ignore

        if self.sparse:
            return csr_array(X, dtype=dtype)
        else:
            return np.array(X, dtype=dtype)



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class AvalonFingerprint(BaseFingerprintTransformer):
    """
    Avalon fingerprint.

    The implementation uses RDKit and Avalon Toolkit [1]_. This is a hashed fingerprint, where
    fragments are computed based on atom environments. The fingerprint is based on
    multiple features, including atom types, bond types, rings and paths. The detailed description
    can be found in the original paper [2]_.

    Parameters
    ----------
    fp_size : int, default=512
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Avalon toolkit
        <https://sourceforge.net/projects/avalontoolkit/>`_

    .. [2] `Gedeck, Peter, Bernhard Rohde, and Christian Bartels
        "QSAR − How Good Is It in Practice? Comparison of Descriptor Sets on an Unbiased
        Cross Section of Corporate Data Sets."
        J. Chem. Inf. Model. 2006, 46, 5, 1924–1936
        <https://pubs.acs.org/doi/abs/10.1021/ci050413p>`_

    Examples
    --------
    >>> from skfp.fingerprints import AvalonFingerprint
    >>> smiles = ["CCO", "CCN"]
    >>> fp = AvalonFingerprint()
    >>> fp
    AvalonFingerprint()
    >>> X = fp.transform(smiles)
    >>> X  # doctest: +ELLIPSIS
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
    }

    def __init__(
        self,
        fp_size: int = 512,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute Avalon fingerprints.

        Parameters
        ----------
        X : {sequence of str or Mol}
            Sequence containing SMILES strings or RDKit molecules.

        copy : bool, default=False
            Whether to copy input data.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.n_features_out)
            Transformed data.
        """
        return super().transform(X, copy=copy)

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Avalon.pyAvalonTools import GetAvalonCountFP, GetAvalonFP

        X = ensure_mols(X)

        if self.count:
            X = [GetAvalonCountFP(mol, nBits=self.fp_size).ToList() for mol in X]
        else:
            X = [GetAvalonFP(mol, nBits=self.fp_size) for mol in X]

        dtype = np.uint32 if self.count else np.uint8

        if self.sparse:
            return csr_array(X, dtype=dtype)
        else:
            return np.array(X, dtype=dtype)



from typing import Union

from rdkit.Chem import GetFormalCharge, Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import (
    CalcNumHBA,
    CalcNumHBD,
    CalcNumHeteroatoms,
    CalcNumRings,
    CalcNumRotatableBonds,
    CalcTPSA,
)

from skfp.bases.base_filter import BaseFilter

from .utils import (
    get_max_ring_size,
    get_non_carbon_to_carbon_ratio,
    get_num_carbon_atoms,
    get_num_charged_functional_groups,
    get_num_rigid_bonds,
)


class FAF4DruglikeFilter(BaseFilter):
    """
    FAFDrugs4 Drug-Like Soft filter.

    Designed as a part of FAFDrugs4 software [1]_ [2]_. Based on literature describing
    physico-chemical properties of drugs and their statistical analysis. Selected
    so that up to 90% of the 916 FDA-approved oral drugs fulfill the rules of this
    filter.

    Known to reject accepted, but atypical drugs, mostly due to high molecular weight
    (e.g. Rinfampin), hydrophobicity (e.g. Probucol), rotatable bonds (e.g. Aliskerin)
    or HBD (e.g. Kanamycin).

    Molecule must fulfill conditions:

    - molecular weight in range ``[100, 600]``
    - logP in range ``[-3, 6]``
    - HBA <= 12
    - HBD <= 7
    - TPSA <= 180
    - number of rotatable bonds <= 11
    - number of rigid bonds <= 30
    - number of rings <= 6
    - max ring size <= 18
    - number of carbons in range ``[3, 35]``
    - number of heteroatoms in range ``[1, 15]``
    - non-carbons to carbons ratio in range ``[0.1, 1.1]``
    - number of charged functional groups <= 4
    - total formal charge in range ``[-4, 4]``

    Note that the FAFDrugs4 uses ChemAxon for determining functional groups. We use
    their publicly available CXSMARTS list of functional groups [3]_. Phosphine and
    sulfoxide patterns could not be parsed by RDKit, so we manually fixed them.

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Details of physico-chemical property filters available in FAF-Drugs4
        <https://fafdrugs4.rpbs.univ-paris-diderot.fr/filters.html>`_

    .. [2] `D. Lagorce et al.
        "FAF-Drugs4: free ADME-tox filtering computations for chemical biology and
        early stages drug discovery"
        Bioinformatics, 33(22), 2017, 3658–3660
        <https://doi.org/10.1093/bioinformatics/btx491>`_

    .. [3] `ChemAxon documentation: Predefined Functional Groups and Named Molecule Groups
        <https://docs.chemaxon.com/display/docs/attachments/attachments_1829721_1_functionalgroups.cxsmi>`_

    Examples
    ----------
    >>> from skfp.filters import FAF4DruglikeFilter
    >>> smiles = ["C", "CC(=O)Nc1ccc(O)cc1"]
    >>> filt = FAF4DruglikeFilter()
    >>> filt
    FAF4DruglikeFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CC(=O)Nc1ccc(O)cc1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation, return_indicators, n_jobs, batch_size, verbose
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            100 <= MolWt(mol) <= 600,
            -3 <= MolLogP(mol) <= 6,
            CalcNumHBA(mol) <= 12,
            CalcNumHBD(mol) <= 7,
            CalcTPSA(mol) <= 180,
            CalcNumRotatableBonds(mol) <= 11,
            get_num_rigid_bonds(mol) <= 30,
            CalcNumRings(mol) <= 6,
            get_max_ring_size(mol) <= 18,
            3 <= get_num_carbon_atoms(mol) <= 35,
            1 <= CalcNumHeteroatoms(mol) <= 15,
            0.1 <= get_non_carbon_to_carbon_ratio(mol) <= 1.1,
            get_num_charged_functional_groups(mol) <= 4,
            -4 <= GetFormalCharge(mol) <= 4,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence

import pandas as pd


def get_data_from_indices(data: Sequence, indices: Sequence[int]) -> list:
    """
    Helper function to retrieve data elements from specified indices.
    """
    if isinstance(data, (pd.Series, pd.DataFrame)):
        return [data.iloc[idx] for idx in set(indices)]
    else:
        return [data[idx] for idx in set(indices)]



from typing import Union

from rdkit.Chem import Crippen, Mol, rdMolDescriptors

from skfp.bases.base_filter import BaseFilter


class GSKFilter(BaseFilter):
    # flake8: noqa E501
    """
    GSK rule (4/400) filter.

    Compute GSK Rule (4/400) for druglikeness using interpretable ADMET rule of thumb [1]_.

    Molecule must fulfill conditions:

    - molecular weight <= 400
    - logP <= 4

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Glesson, M. P.
        "Generation of a Set of Simple, Interpretable ADMET Rules of Thumb"
        J. Med. Chem. 2008, 51, 4, 817–834
        <https://pubs.acs.org/doi/10.1021/jm701122q>`_

    Examples
    ----------
    >>> from skfp.filters import GSKFilter
    >>> smiles = ["C1CC1N2C=C(C(=O)C3=CC(=C(C=C32)N4CCNCC4)F)C(=O)O", "O=C(O)c1ccccc1c2ccc(cc2)Cn3c4cc(cc(c4nc3CCC)C)c5nc6ccccc6n5C"]
    >>> filt = GSKFilter()
    >>> filt
    GSKFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C1CC1N2C=C(C(=O)C3=CC(=C(C=C32)N4CCNCC4)F)C(=O)O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation, return_indicators, n_jobs, batch_size, verbose
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            rdMolDescriptors.CalcExactMolWt(mol) <= 400,
            Crippen.MolLogP(mol) <= 4,
        ]

        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class RDKitFingerprint(BaseFingerprintTransformer):
    """
    RDKit fingerprint.

    This fingerprint is an RDKit original [1]_. This is a hashed fingerprint,
    where fragments are created from small subgraphs on the molecular graph.

    For a given molecule, all paths between ``min_path`` and ``max_path`` (inclusive)
    are extracted and hashed, based on bond invariants (see below). Those are any
    subgraphs, unless ``linear_paths_only`` is set to True. Note that all explicit
    atoms, including hydrogens if present, are used.

    Each subgraph is hashed. Based on this hash value, ``nBitsPerHash`` pseudorandom
    numbers are generated and used to set bits in the resulting fingerprint. Finally,
    it is folded to ``fp_size`` length.

    Subgraphs are identified based on bonds constituting them. Bonds invariants (types,
    features) take into consideration:

    - atomic numbers and aromaticity of bonded atoms
    - degrees of bonded atoms
    - bond type/order (single, double, triple, aromatic)

    For more details on fingerprints of this type, see Daylight documentation [2]_.

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    min_path : int, default=1
        Minimal length of paths used, in bonds. Default value means that at least
        2-atom subgraphs are used.

    max_path : int, default=7
        Maximal length of paths used, in bonds.

    use_pharmacophoric_invariants: bool, default=False
        Whether to use pharmacophoric invariants (atom types) instead of default ones.
        They are the same as in the FCFP fingerprint: Donor, Acceptor, Aromatic,
        Halogen, Basic, Acidic.

    use_bond_order : bool, default=True
        Whether to take bond order (type) into consideration when hashing subgraphs.
        False means that only graph topology (subgraph shape) is used.

    num_bits_per_feature : int, default=2
        How many bits to set for each subgraph.

    linear_paths_only : bool, default=False
        Whether to use only linear paths, instead of any subgraphs.

    count_simulation : bool, default=True
        Whether to use count simulation for approximating feature counts [3]_.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`LayeredFingerprint` : Related fingerprint, but uses different atom and bond
        types to set multiple bits.

    References
    ----------
    .. [1] `Gregory Landrum
        "Fingerprints in the RDKit"
        UGM 2012
        <https://www.rdkit.org/UGM/2012/Landrum_RDKit_UGM.Fingerprints.Final.pptx.pdf>`_

    .. [2] `Daylight documentation
        "Fingerprints - Screening and Similarity"
        <https://www.daylight.com/dayhtml/doc/theory/theory.finger.html>`_

    .. [3] `Greg Landrum
        "Simulating count fingerprints"
        <https://greglandrum.github.io/rdkit-blog/posts/2021-07-06-simulating-counts.html>`_

    Examples
    --------
    >>> from skfp.fingerprints import RDKitFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = RDKitFingerprint()
    >>> fp
    RDKitFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "min_path": [Interval(Integral, 1, None, closed="left")],
        "max_path": [Interval(Integral, 1, None, closed="left")],
        "use_bond_order": ["boolean"],
        "num_bits_per_feature": [Interval(Integral, 1, None, closed="left")],
        "linear_paths_only": ["boolean"],
        "count_simulation": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        min_path: int = 1,
        max_path: int = 7,
        use_pharmacophoric_invariants: bool = False,
        use_bond_order: bool = True,
        num_bits_per_feature: int = 2,
        linear_paths_only: bool = False,
        count_simulation: bool = False,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.min_path = min_path
        self.max_path = max_path
        self.use_pharmacophoric_invariants = use_pharmacophoric_invariants
        self.use_bond_order = use_bond_order
        self.num_bits_per_feature = num_bits_per_feature
        self.linear_paths_only = linear_paths_only
        self.count_simulation = count_simulation

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.max_path < self.min_path:
            raise InvalidParameterError(
                f"The max_path parameter of {self.__class__.__name__} must be "
                f"greater or equal to min_path, got: "
                f"min_path={self.min_path}, max_path={self.max_path}"
            )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdFingerprintGenerator import (
            GetMorganFeatureAtomInvGen,
            GetRDKitFPGenerator,
        )

        X = ensure_mols(X)

        if self.use_pharmacophoric_invariants:
            inv_gen = GetMorganFeatureAtomInvGen()
        else:
            inv_gen = None

        gen = GetRDKitFPGenerator(
            fpSize=self.fp_size,
            minPath=self.min_path,
            maxPath=self.max_path,
            atomInvariantsGenerator=inv_gen,
            useBondOrder=self.use_bond_order,
            numBitsPerFeature=self.num_bits_per_feature,
            branchedPaths=not self.linear_paths_only,
            countSimulation=self.count_simulation,
        )

        if self.count:
            X = [gen.GetCountFingerprintAsNumPy(mol) for mol in X]
        else:
            X = [gen.GetFingerprintAsNumPy(mol) for mol in X]

        dtype = np.uint32 if self.count else np.uint8
        if self.sparse:
            return csr_array(X, dtype=dtype)
        else:
            return np.array(X, dtype=dtype)



from collections.abc import Sequence
from numbers import Real
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids


class WHIMFingerprint(BaseFingerprintTransformer):
    """
    WHIM (Weighted Holistic Invariant Molecular descriptors) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, where
    bits measure rotation-invariant 3D information regarding size, shape, symmetry
    and atom distributions.

    Features are based on the principal component analysis (PCA) on the centered
    cartesian coordinates of a molecule by using a weighted covariance matrix. There are
    two grups of features, each one measuring size, shape, symmetry and density of atoms:
    - 11 directional, using scores of individual principal axes
    - 7 global, aggregating information about the whole molecule

    Additionally, all directional features and 5 of the global ones are computed
    using unweighted distances matrix, as well as 6 weighted variants, using atomic
    features: atomic mass, van der Waals volume, electronegativity, polarizability,
    ion polarity, and IState [1]_ [2]_. They are relative to the carbon, e.g. molecular
    weight is: MW(atom_type) / MW(carbon).

    This gives 114 features: 11 * 7 weighted directional + 5 * 7 weighted global
    + 2 unweighted global. See [3]_ [4]_ [5]_ [6]_ for details.

    Typical correct values should be small, but can result in NaN or infinity for some
    molecules. Value clipping with ``clip_val`` parameter, feature selection, and/or
    imputation should be used.

    Parameters
    ----------
    clip_val : float or None, default=2147483647
        Value to clip results at, both positive and negative ones.The default value is
        the maximal value of 32-bit integer, but should often be set lower, depending
        on the application. ``None`` means that no clipping is applied.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 114
        Number of output features, size of fingerprints.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    References
    ----------
    .. [1] RDKit IState discussion
        https://github.com/rdkit/rdkit/discussions/6122

    .. [2] RDKit IState implementation
        https://github.com/rdkit/rdkit/blob/df2b6c8a8c775748c1dcac83af0f92af434bab81/Code/GraphMol/Descriptors/MolData3Ddescriptors.cpp#L127

    .. [3] `Rajarshi Guha
        "Weighted Holistic Invariant Molecular Descriptors"
        <http://www.rguha.net/writing/notes/desc/node7.html>`_

    .. [4] `Roberto Todeschini and Viviana Consonni
        "Molecular Descriptors for Chemoinformatics"
        <https://onlinelibrary.wiley.com/doi/book/10.1002/9783527628766>`_

    .. [5] `Guillaume Godin
        "3D	descriptors in RDKit"
        UGM 2017
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Godin_3D_Descriptors.pdf>`_

    .. [6] `Roberto Todeschini and Paola Gramatica
        "The WHIM Theory: New 3D Molecular Descriptors for QSAR in Environmental Modelling"
        SAR and QSAR in Environmental Research, 7(1–4), 89–115
        <https://www.tandfonline.com/doi/abs/10.1080/10629369708039126>`_

    Examples
    --------
    >>> from skfp.fingerprints import WHIMFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = WHIMFingerprint()
    >>> fp
    WHIMFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)  # doctest: +SKIP
    array([[0.44 , 0.067, 0.   , ..., 0.514, 0.537, 0.537],
           [1.17 , 0.395, 0.393, ..., 2.266, 3.38 , 2.542],
           [0.329, 0.   , 0.   , ..., 0.329, 0.329, 0.329],
           [1.196, 0.507, 0.242, ..., 2.183, 3.285, 3.129]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "clip_val": [None, Interval(Real, 0, None, closed="left")],
    }

    def __init__(
        self,
        clip_val: float = 2147483647,  # max int32 value
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=114,
            requires_conformers=True,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.clip_val = clip_val

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import CalcWHIM

        X = require_mols_with_conf_ids(X)
        X = [CalcWHIM(mol, confId=mol.GetIntProp("conf_id")) for mol in X]
        if self.clip_val is not None:
            X = np.clip(X, -self.clip_val, self.clip_val)

        return csr_array(X) if self.sparse else np.array(X)



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP, MolMR
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import CalcNumAtoms

from skfp.bases.base_filter import BaseFilter


class GhoseFilter(BaseFilter):
    """
    Ghose Filter.

    Used to searching for drug-like molecules [1]_.

    Molecule must fulfill conditions:

        - 160 <= molecular weight <= 400
        - -0.4 <= logP <= 5.6
        - 20 <= number of atoms <= 70
        - 40 <= refractivity <= 130

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Ghose, A. K., Viswanadhan, V. N., & Wendoloski, J. J.
        "A Knowledge-Based Approach in Designing Combinatorial or Medicinal Chemistry Libraries for Drug Discovery. 1.
        A Qualitative and Quantitative Characterization of Known Drug Databases."
        Journal of Combinatorial Chemistry, 1(1), 55–68.
        <https://doi.org/10.1021/cc9800071>`_

    Examples
    ----------
    >>> from skfp.filters import GhoseFilter
    >>> smiles = ["CC(=O)C1=C(O)C(=O)N(CCc2c[nH]c3ccccc23)C1c1ccc(C)cc1", "CC(=O)c1c(C)n(CC2CCCO2)c2ccc(O)cc12",\
    "CC(=O)c1c(C(C)=O)c(C)n(CCCCn2c(C)c(C(C)=O)c(C(C)=O)c2C)c1C"]
    >>> filt = GhoseFilter()
    >>> filt
    GhoseFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CC(=O)C1=C(O)C(=O)N(CCc2c[nH]c3ccccc23)C1c1ccc(C)cc1', 'CC(=O)c1c(C)n(CC2CCCO2)c2ccc(O)cc12']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            160 <= MolWt(mol) <= 400,
            40 <= MolMR(mol) <= 130,
            20 <= CalcNumAtoms(mol) <= 70,
            -0.4 <= MolLogP(mol) <= 5.6,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class LayeredFingerprint(BaseFingerprintTransformer):
    """
    Layered fingerprint.

    This fingerprint is an RDKit original [1]_. This is a hashed fingerprint,
    where fragments are created from small subgraphs on the molecular graph.

    For a given molecule, all paths between ``min_path`` and ``max_path`` (inclusive)
    are extracted. Those are any subgraphs, unless ``linear_paths_only`` is set to True.
    Note that all explicit atoms, including hydrogens if present, are used.

    Then each subgraph is hashed in "layers" (hence the fingerprint name), using
    different atom and bond invariants (features). Additionally, information about path
    length and the number of distinct atoms is included. Those hashes are combined into
    a single value and hashed into the resulting fingerprint, which is folded at the end.

    Layers are:
    - pure topology (using only subgraph "shape")
    - bond order (type), ignoring aromaticiyt (aromatic bonds are treated as single)
    - atom types (atomic numbers)
    - presence of rings (whether bond is in a ring)
    - ring sizes (size of smallest ring that bond is a part of)
    - aromaticity (whether bond is aromatic or not)

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    min_path : int, default=1
        Minimal length of paths used, in bonds. Default value means that at least
        2-atom subgraphs are used.

    max_path : int, default=7
        Maximal length of paths used, in bonds.

    linear_paths_only : bool, default=False
        Whether to use only linear paths, instead of any subgraphs.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`RDKitFingerprint` : Related fingerprint, but uses constant bond types
        and pseudorandom numbers to set multiple bits.

    References
    ----------
    .. [1] `Gregory Landrum
        "Fingerprints in the RDKit"
        UGM 2012
        <https://www.rdkit.org/UGM/2012/Landrum_RDKit_UGM.Fingerprints.Final.pptx.pdf>`_

    Examples
    --------
    >>> from skfp.fingerprints import LayeredFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = LayeredFingerprint()
    >>> fp
    LayeredFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "min_path": [Interval(Integral, 1, None, closed="left")],
        "max_path": [Interval(Integral, 1, None, closed="left")],
        "linear_paths_only": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        min_path: int = 1,
        max_path: int = 7,
        linear_paths_only: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.min_path = min_path
        self.max_path = max_path
        self.linear_paths_only = linear_paths_only

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.max_path < self.min_path:
            raise InvalidParameterError(
                f"The max_path parameter of {self.__class__.__name__} must be "
                f"greater or equal to min_path, got: "
                f"min_path={self.min_path}, max_path={self.max_path}"
            )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdmolops import LayeredFingerprint as RDKitLayeredFingerprint

        X = ensure_mols(X)
        X = [
            RDKitLayeredFingerprint(
                mol,
                fpSize=self.fp_size,
                minPath=self.min_path,
                maxPath=self.max_path,
                branchedPaths=not self.linear_paths_only,
            )
            for mol in X
        ]

        if self.sparse:
            return csr_array(X, dtype=np.uint8)
        else:
            return np.array(X, dtype=np.uint8)



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class EStateFingerprint(BaseFingerprintTransformer):
    """
    Electrotopological State (EState) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, where
    bits measure contributions of different atom types, based on their electronic
    and topological properties. Based on ``variant`` argument, either counts of
    different atom types, or sums of their properties are computed.

    The idea is to represent the intrinsic electronic state of the atom, in the context
    of the particular molecule, i.e. taking into consideration the electronic influence
    of all other atoms and the molecule topology (structure).

    79 atom types are used, as defined in the original paper [1]_. For practical
    implementation, they are formulated as SMARTS patterns, selecting individual
    atoms of particular type [2]_. Generally, they take into consideration:

    - atom element
    - valence state (including aromaticity)
    - number of bonded hydrogens
    - in some cases, the identity of other bonded atoms

    Parameters
    ----------
    variant : {"bit", "count", "sum"}, default="sum"
        Fingerprint variant. Default "sum" results in the sum of EState indices
        for each atom type. "count" results in integer vector with counts of atoms
        of particular types, and "bit" simply denotes existence of given atom types.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Lowell H. Hall and Lemont B. Kier
        "Electrotopological State Indices for Atom Types: A Novel Combination of Electronic,
        Topological, and Valence State Information"
        J. Chem. Inf. Comput. Sci. 1995, 35, 6, 1039–1045
        <https://pubs.acs.org/doi/10.1021/ci00028a014>`_

    .. [2] `Gregory Landrum and Rational Discovery LLC
        RDKit - EState Atom Types
        <https://github.com/rdkit/rdkit/blob/master/rdkit/Chem/EState/AtomTypes.py>`_

    Examples
    --------
    >>> from skfp.fingerprints import EStateFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = EStateFingerprint()
    >>> fp
    EStateFingerprint()

    >>> fp.transform(smiles)  # doctest: +ELLIPSIS
    array([[0.        , 0.        , ..., 0.        , 0.        ],
           [0.        , 0.        , ..., 0.        , 0.        ],
           [0.        , 0.        , ..., 0.        , 0.        ],
           [0.        , 0.        , ..., 0.        , 0.        ]])

    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "variant": [StrOptions({"bit", "count", "sum"})],
    }

    def __init__(
        self,
        variant: str = "sum",
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=79,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.variant = variant

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.EState.Fingerprinter import FingerprintMol

        X = ensure_mols(X)

        X = np.array([FingerprintMol(mol) for mol in X])
        if self.variant == "bit":
            X = (X[:, 0] > 0).astype(np.uint8)
        elif self.variant == "count":
            X = (X[:, 0]).astype(np.uint32)
        else:  # "sum" variant
            X = X[:, 1]

        return csr_array(X) if self.sparse else np.array(X)



import os.path
import warnings
from collections.abc import Sequence
from copy import deepcopy

from rdkit.Chem import Mol, SDMolSupplier, SDWriter
from rdkit.Chem.PropertyMol import PropertyMol

from skfp.bases import BasePreprocessor
from skfp.utils.validators import check_mols


class MolFromSDFTransformer(BasePreprocessor):
    """
    Creates RDKit ``Mol`` objects from SDF string or file.

    SDF (structure-data format) is processed for whole files, rather than individual
    molecules. For this reason ``.transform()`` either reads the SDF file directly
    from disk or takes a string input in that format.

    For details see RDKit documentation [1]_.

    Parameters
    ----------
    sanitize : bool, default=True
        Whether to perform sanitization [1]_, i.e. basic validity checks, on created
        molecules.

    remove_hydrogens : bool, default=True
        Remove explicit hydrogens from the molecule where possible, using RDKit
        implicit hydrogens instead.

    References
    ----------
    .. [1] `RDKit SDMolSupplier documentation
        <https://rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.SDMolSupplier>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolFromSDFTransformer
    >>> sdf_file_path = "mols_in.sdf"
    >>> mol_from_sdf = MolFromSDFTransformer()  # doctest: +SKIP
    >>> mol_from_sdf  # doctest: +SKIP
    MolFromSDFTransformer()

    >>> mol_from_sdf.transform(sdf_file_path)  # doctest: +SKIP
        [<rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>]
    """

    _parameter_constraints: dict = {
        "sanitize": ["boolean"],
        "remove_hydrogens": ["boolean"],
    }

    def __init__(
        self,
        sanitize: bool = True,
        remove_hydrogens: bool = True,
    ):
        super().__init__()
        self.sanitize = sanitize
        self.remove_hydrogens = remove_hydrogens

    def transform(self, X: str, copy: bool = False) -> list[Mol]:
        """
        Create RDKit ``Mol`` objects from SDF file.

        Parameters
        ----------
        X : str
            Path to SDF file.

        copy : bool, default=False
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,)
            List with RDKit ``Mol`` objects.
        """
        self._validate_params()

        if X.endswith(".sdf"):
            if not os.path.exists(X):
                raise FileNotFoundError(f"SDF file at path '{X}' not found")

            with open(X) as file:
                X = file.read()

        supplier = SDMolSupplier()
        supplier.SetData(X, sanitize=self.sanitize, removeHs=self.remove_hydrogens)
        mols = list(supplier)

        if not mols:
            warnings.warn("No molecules detected in provided SDF file")

        return mols


class MolToSDFTransformer(BasePreprocessor):
    """
    Creates SDF file from RDKit ``Mol`` objects.

    SDF (structure-data format) is processed for whole files, rather than individual
    molecules. For this reason ``.transform()`` saves the results directly to file.

    If ``conf_id`` integer property is set for molecules, they are used to determine
    the conformer to save.

    For details see RDKit documentation [1]_.

    Parameters
    ----------
    filepath : string, default="mols.sdf"
        A string with file path location to save the SDF file. Tt should be a valid
        file path with ``.sdf`` extention.

    kekulize : bool, default=True
        Whether to kekulize molecules before writing them to SDF file.

    force_V3000 : bool, default=False
        Whether to force the V3000 format when writing to SDF file.

    References
    ----------
    .. [1] `RDKit SDWriter documentation
        <https://rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.SDWriter>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolFromSDFTransformer, MolToSDFTransformer
    >>> sdf_file_path = "mols_in.sdf"
    >>> mol_from_sdf = MolFromSDFTransformer()
    >>> mol_to_sdf = MolToSDFTransformer(filepath="mols_out.sdf")
    >>> mol_to_sdf
    MolToSDFTransformer(filepath='mols_out.sdf')

    >>> mols = mol_from_sdf.transform(sdf_file_path)  # doctest: +SKIP
    >>> mol_to_sdf.transform(mols)  # doctest: +SKIP
    """

    _parameter_constraints: dict = {
        "filepath": [str],
        "kekulize": ["boolean"],
        "force_V3000": ["boolean"],
    }

    def __init__(
        self,
        filepath: str = "mols.sdf",
        kekulize: bool = True,
        force_V3000: bool = False,
    ):
        super().__init__()
        self.filepath = filepath
        self.kekulize = kekulize
        self.force_V3000 = force_V3000

    def transform(self, X: Sequence[Mol], copy: bool = False) -> None:
        self._validate_params()
        check_mols(X)

        if copy:
            X = deepcopy(X)

        with open(self.filepath, "w") as file:
            writer = SDWriter(file)
            writer.SetKekulize(self.kekulize)
            writer.SetForceV3000(self.force_V3000)

            for mol in X:
                if isinstance(mol, PropertyMol) and mol.HasProp("conf_id"):
                    writer.write(mol, confId=mol.GetIntProp("conf_id"))
                else:
                    writer.write(mol)

            writer.flush()



from collections.abc import Sequence
from copy import deepcopy
from numbers import Real
from typing import Optional, Union

import numpy as np
from numpy.linalg import norm
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from scipy.stats import moment
from sklearn.utils._param_validation import Interval, StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids


class ElectroShapeFingerprint(BaseFingerprintTransformer):
    """
    ElectroShape fingerprint.

    This is a descriptor-based fingerprint, extending the USR fingerprint by
    additionally considering atomic partial charges [1]_.

    It first computes atomic partial charges, and then uses both conformational
    (spatial) structure, and this electric information, to compute reference
    points (centroids). First three are like in USR, and last two
    additionally use partial charge in distance calculation. See the original paper
    [1]_ for details. For each centroid, the distribution of distances between atoms
    and the centroid is aggregated using the first three moments (mean, standard
    deviation, cubic root of skewness). This results in 15 features.

    This is a 3D fingerprint, and requries molecules with ``conf_id`` integer property
    set. They can be generated with :class:`~skfp.preprocessing.ConformerGenerator`.
    Furthermore, only molecules with 3 or more atoms are allowed, to allow computation
    of all three moments.

    Typical correct values should be small, but problematic molecules may result in NaN
    values for some descriptors. In those cases, imputation should be used.

    Parameters
    ----------
    partial_charge_model : {"Gasteiger", "MMFF94", "formal", "precomputed"}, default="formal"
        Which model to use to compute atomic partial charges. Default ``"formal"``
        computes formal charges, and is the simplest and most error-resistantone.
        ``"precomputed"`` assumes that the inputs are RDKit PropertyMol objects
        with "charge" float property set.

    charge_scaling_factor : float, default=25.0
        Partial charges are multiplied by this factor to bring them to a value
        range comparable to distances in Angstroms.

    charge_errors : {"raise", "ignore", "zero"}, default="raise"
        How to handle errors during calculation of atomic partial charges. ``"raise"``
        immediately raises any errors. ``"NaN"`` ignores any atoms that failed the
        computation; note that if all atoms fail, the error will be raised (use
        ``errors`` parameter to control this). ``"zero"`` uses default value of 0 to
        fill all problematic charges.

    errors : {"raise", "NaN", "ignore"}, default="raise"
        How to handle errors during fingerprint calculation. ``"raise"`` immediately
        raises any errors. ``"NaN"`` returns NaN values for molecules which resulted in
        errors. ``"ignore"`` suppresses errors and does not return anything for
        molecules with errors. This potentially results in less output vectors than
        input molecules, and should be used with caution.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 15
        Number of output features, size of fingerprints.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    See Also
    --------
    :class:`USR` : Related fingerprint, which ElectroShape expands.

    :class:`USRCAT` : Related fingerprint, which expands USR with pharmacophoric
        atom types, instead of partial charges.

    References
    ----------
    .. [1] `Armstrong, M.S., Morris, G.M., Finn, P.W. et al.
        "ElectroShape: fast molecular similarity calculations incorporating shape, chirality and electrostatics"
        J Comput Aided Mol Des 24, 789–801 (2010)
        <https://doi.org/10.1007/s10822-010-9374-0>`_

    Examples
    --------
    >>> from skfp.fingerprints import ElectroShapeFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["CC=O"]
    >>> fp = ElectroShapeFingerprint()
    >>> fp
    ElectroShapeFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)  # doctest: +SKIP
    array([[ 4.84903774,  5.10822298, ...        ,  5.14008906,  2.75483277 ]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "partial_charge_model": [
            StrOptions({"Gasteiger", "MMFF94", "formal", "precomputed"})
        ],
        "charge_scaling_factor": [Interval(Real, 0.0, None, closed="neither")],
        "charge_errors": [StrOptions({"raise", "ignore", "zero"})],
        "errors": [StrOptions({"raise", "NaN", "ignore"})],
    }

    def __init__(
        self,
        partial_charge_model: str = "formal",
        charge_scaling_factor: float = 25.0,
        charge_errors: str = "raise",
        errors: str = "raise",
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=15,
            requires_conformers=True,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.partial_charge_model = partial_charge_model
        self.charge_scaling_factor = charge_scaling_factor
        self.charge_errors = charge_errors
        self.errors = errors

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute ElectroShape fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects, with conformers generated and
            ``conf_id`` integer property set.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.
        """
        y = np.empty(len(X))
        X, _ = self.transform_x_y(X, y, copy=copy)
        return X

    def transform_x_y(
        self, X: Sequence[Mol], y: np.ndarray, copy: bool = False
    ) -> tuple[Union[np.ndarray, csr_array], np.ndarray]:
        """
        Compute ElectroShape fingerprints. The returned values for X and y are
        properly synchronized.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects, with conformers generated and
            ``conf_id`` integer property set.

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.

        copy : bool, default=False
            Copy the inputs X and y or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.
        """
        if copy:
            X = deepcopy(X)
            y = deepcopy(y)

        X = super().transform(X)

        if self.errors == "ignore":
            # errors are marked as NaN rows
            idxs_to_keep = [idx for idx, x in enumerate(X) if not np.any(np.isnan(x))]
            X = X[idxs_to_keep]
            y = y[idxs_to_keep]

        return X, y

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        mols = require_mols_with_conf_ids(X)

        if self.errors == "raise":
            fps = [self._get_fp(mol) for mol in mols]
        else:  # self.errors in {"NaN", "ignore"}
            fps = []
            for mol in X:
                try:
                    fp = self._get_fp(mol)
                except ValueError:
                    fp = np.full(self.n_features_out, np.NaN)
                fps.append(fp)

        return np.array(fps)

    def _get_fp(self, mol: Mol) -> np.ndarray:
        conf_id = mol.GetIntProp("conf_id")
        coords = mol.GetConformer(conf_id).GetPositions()
        charges = self._get_atomic_charges(mol) * self.charge_scaling_factor
        descriptors = np.column_stack((coords, charges))
        centroid_dists = self._get_centroid_distances(descriptors, charges)

        fp = []
        for d in centroid_dists:
            # three moments: mean, stddev, cubic root of skewness
            fp.extend([np.mean(d), np.std(d), np.cbrt(moment(d, 3))])

        return np.array(fp)

    def _get_atomic_charges(self, mol: Mol) -> np.ndarray:
        from rdkit.Chem import MolToSmiles, rdPartialCharges
        from rdkit.Chem.rdForceFieldHelpers import MMFFGetMoleculeProperties

        atoms = mol.GetAtoms()

        if self.partial_charge_model == "Gasteiger":
            rdPartialCharges.ComputeGasteigerCharges(mol)
            charges = [atom.GetDoubleProp("_GasteigerCharge") for atom in atoms]
        elif self.partial_charge_model == "MMFF94":
            values = MMFFGetMoleculeProperties(mol)
            charges = [
                values.GetMMFFPartialCharge(i) if values else None
                for i in range(len(atoms))
            ]
        elif self.partial_charge_model == "formal":
            charges = [atom.GetFormalCharge() for atom in atoms]
        else:  # precomputed
            charges = [atom.GetDoubleProp("charge") for atom in atoms]

        if self.charge_errors == "raise" and any(charge is None for charge in charges):
            smiles = MolToSmiles(mol)
            raise ValueError(
                f"Failed to compute at least one atom partial charge for {smiles}"
            )
        elif self.charge_errors == "zero":
            charges = np.nan_to_num(np.array(charges, dtype=float), nan=0)
        else:  # "ignore"
            charges = [c for c in charges if c is not None]

        return np.asarray(charges, dtype=float)

    def _get_centroid_distances(
        self, descriptors: np.ndarray, charges: np.ndarray
    ) -> list[np.ndarray]:
        # geometric center
        c1 = descriptors.mean(axis=0)
        dists_c1 = norm(descriptors - c1, axis=1)

        # furthest atom from c1
        c2 = descriptors[dists_c1.argmax()]
        dists_c2 = norm(descriptors - c2, axis=1)

        # furthest atom from c2
        c3 = descriptors[dists_c2.argmax()]
        dists_c3 = norm(descriptors - c3, axis=1)

        # vectors between centroids
        vec_a = c2 - c1
        vec_b = c3 - c1

        # scaled vector product of spatial coordinates (a_S and b_S)
        # it distinguishes between a chiral molecule and its enantiomer
        cross_ab = np.cross(vec_a[:3], vec_b[:3])
        cross_ab_norm = norm(cross_ab)
        if np.isclose(cross_ab_norm, 0):
            vec_c = 0
        else:
            vec_c = (norm(vec_a) / (2 * cross_ab_norm)) * cross_ab

        # geometric mean centroid moved in the direction of smallest and largest charge
        # note that charges were already scaled before
        c4 = np.append(c1[:3] + vec_c, np.max(charges))
        c5 = np.append(c1[:3] + vec_c, np.min(charges))

        dists_c4 = norm(descriptors - c4, axis=1)
        dists_c5 = norm(descriptors - c5, axis=1)

        return [dists_c1, dists_c2, dists_c3, dists_c4, dists_c5]



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import rdkit.Chem.Fragments
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class FunctionalGroupsFingerprint(BaseFingerprintTransformer):
    """
    Functional groups fingerprint.

    The implementation uses RDKit. This is a substructure, descriptor fingerprint,
    checking occurrences of 85 functional groups (also known as fragments) available
    in RDKit [1]_.

    Parameters
    ----------
    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 85
        Number of output features, size of fingerprints.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] RDKit, rdkit.Chem.Fragments module
        https://www.rdkit.org/docs/source/rdkit.Chem.Fragments.html

    Examples
    --------
    >>> from skfp.fingerprints import FunctionalGroupsFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O", "CCO", "CCN"]
    >>> fp = FunctionalGroupsFingerprint()
    >>> fp
    FunctionalGroupsFingerprint()
    >>> X = fp.transform(smiles)
    >>> X
    array([[0, 0, 0, ..., 1, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 1, 0, 0],
           [0, 0, 0, ..., 1, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    def __init__(
        self,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=85,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        X = ensure_mols(X)

        X = np.array([self._get_functional_groups_counts(mol) for mol in X])

        if not self.count:
            X = X > 0

        dtype = np.uint32 if self.count else np.uint8
        return csr_array(X, dtype=dtype) if self.sparse else X.astype(dtype)

    def _get_functional_groups_counts(self, mol: Mol) -> list[int]:
        func_groups_counts = [
            func_group(mol)
            for func_group in [
                rdkit.Chem.Fragments.fr_Al_COO,
                rdkit.Chem.Fragments.fr_Al_OH,
                rdkit.Chem.Fragments.fr_Al_OH_noTert,
                rdkit.Chem.Fragments.fr_ArN,
                rdkit.Chem.Fragments.fr_Ar_COO,
                rdkit.Chem.Fragments.fr_Ar_N,
                rdkit.Chem.Fragments.fr_Ar_NH,
                rdkit.Chem.Fragments.fr_Ar_OH,
                rdkit.Chem.Fragments.fr_COO,
                rdkit.Chem.Fragments.fr_COO2,
                rdkit.Chem.Fragments.fr_C_O,
                rdkit.Chem.Fragments.fr_C_O_noCOO,
                rdkit.Chem.Fragments.fr_C_S,
                rdkit.Chem.Fragments.fr_HOCCN,
                rdkit.Chem.Fragments.fr_Imine,
                rdkit.Chem.Fragments.fr_NH0,
                rdkit.Chem.Fragments.fr_NH1,
                rdkit.Chem.Fragments.fr_NH2,
                rdkit.Chem.Fragments.fr_N_O,
                rdkit.Chem.Fragments.fr_Ndealkylation1,
                rdkit.Chem.Fragments.fr_Ndealkylation2,
                rdkit.Chem.Fragments.fr_Nhpyrrole,
                rdkit.Chem.Fragments.fr_SH,
                rdkit.Chem.Fragments.fr_aldehyde,
                rdkit.Chem.Fragments.fr_alkyl_carbamate,
                rdkit.Chem.Fragments.fr_alkyl_halide,
                rdkit.Chem.Fragments.fr_allylic_oxid,
                rdkit.Chem.Fragments.fr_amide,
                rdkit.Chem.Fragments.fr_amidine,
                rdkit.Chem.Fragments.fr_aniline,
                rdkit.Chem.Fragments.fr_aryl_methyl,
                rdkit.Chem.Fragments.fr_azide,
                rdkit.Chem.Fragments.fr_azo,
                rdkit.Chem.Fragments.fr_barbitur,
                rdkit.Chem.Fragments.fr_benzene,
                rdkit.Chem.Fragments.fr_benzodiazepine,
                rdkit.Chem.Fragments.fr_bicyclic,
                rdkit.Chem.Fragments.fr_diazo,
                rdkit.Chem.Fragments.fr_dihydropyridine,
                rdkit.Chem.Fragments.fr_epoxide,
                rdkit.Chem.Fragments.fr_ester,
                rdkit.Chem.Fragments.fr_ether,
                rdkit.Chem.Fragments.fr_furan,
                rdkit.Chem.Fragments.fr_guanido,
                rdkit.Chem.Fragments.fr_halogen,
                rdkit.Chem.Fragments.fr_hdrzine,
                rdkit.Chem.Fragments.fr_hdrzone,
                rdkit.Chem.Fragments.fr_imidazole,
                rdkit.Chem.Fragments.fr_imide,
                rdkit.Chem.Fragments.fr_isocyan,
                rdkit.Chem.Fragments.fr_isothiocyan,
                rdkit.Chem.Fragments.fr_ketone,
                rdkit.Chem.Fragments.fr_ketone_Topliss,
                rdkit.Chem.Fragments.fr_lactam,
                rdkit.Chem.Fragments.fr_lactone,
                rdkit.Chem.Fragments.fr_methoxy,
                rdkit.Chem.Fragments.fr_morpholine,
                rdkit.Chem.Fragments.fr_nitrile,
                rdkit.Chem.Fragments.fr_nitro,
                rdkit.Chem.Fragments.fr_nitro_arom,
                rdkit.Chem.Fragments.fr_nitro_arom_nonortho,
                rdkit.Chem.Fragments.fr_nitroso,
                rdkit.Chem.Fragments.fr_oxazole,
                rdkit.Chem.Fragments.fr_oxime,
                rdkit.Chem.Fragments.fr_para_hydroxylation,
                rdkit.Chem.Fragments.fr_phenol,
                rdkit.Chem.Fragments.fr_phenol_noOrthoHbond,
                rdkit.Chem.Fragments.fr_phos_acid,
                rdkit.Chem.Fragments.fr_phos_ester,
                rdkit.Chem.Fragments.fr_piperdine,
                rdkit.Chem.Fragments.fr_piperzine,
                rdkit.Chem.Fragments.fr_priamide,
                rdkit.Chem.Fragments.fr_prisulfonamd,
                rdkit.Chem.Fragments.fr_pyridine,
                rdkit.Chem.Fragments.fr_quatN,
                rdkit.Chem.Fragments.fr_sulfide,
                rdkit.Chem.Fragments.fr_sulfonamd,
                rdkit.Chem.Fragments.fr_sulfone,
                rdkit.Chem.Fragments.fr_term_acetylene,
                rdkit.Chem.Fragments.fr_tetrazole,
                rdkit.Chem.Fragments.fr_thiazole,
                rdkit.Chem.Fragments.fr_thiocyan,
                rdkit.Chem.Fragments.fr_thiophene,
                rdkit.Chem.Fragments.fr_unbrch_alkane,
                rdkit.Chem.Fragments.fr_urea,
            ]
        ]
        return func_groups_counts



import json
import os
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
from huggingface_hub import snapshot_download
from huggingface_hub.utils import (
    are_progress_bars_disabled,
    disable_progress_bars,
    enable_progress_bars,
)
from sklearn.datasets import get_data_home as get_sklearn_data_home


def fetch_dataset(
    data_dir: Optional[Union[str, os.PathLike]],
    dataset_name: str,
    filename: str,
    verbose: bool = False,
) -> pd.DataFrame:
    """
    Fetch the dataset from HuggingFace Hub. Returns loaded DataFrame.
    """
    data_home_dir = get_data_home_dir(data_dir, dataset_name)
    dataset_dir = hf_hub_download(data_home_dir, dataset_name, verbose)
    filepath = Path(dataset_dir) / filename
    return pd.read_csv(filepath)


def fetch_splits(
    data_dir: Optional[Union[str, os.PathLike]],
    dataset_name: str,
    filename: str,
    verbose: bool = False,
) -> dict[str, list[int]]:
    """
    Fetch the dataset splits from HuggingFace Hub. Returns loaded JSON with split
    names as keys (e.g. train, valid, test), and lists of indexes as values.
    """
    data_home_dir = get_data_home_dir(data_dir, dataset_name)
    dataset_dir = hf_hub_download(data_home_dir, dataset_name, verbose)
    filepath = Path(dataset_dir) / filename
    if verbose:
        print(filepath)
    with open(filepath) as file:
        return json.load(file)


def get_data_home_dir(
    data_dir: Optional[Union[str, os.PathLike]], dataset_name: str
) -> str:
    """
    Get the data home directory. If valid dataset path is provided, first
    ensures it exists (and all directories in the path). Otherwise, it uses the
    scikit-learn directory, by default `$HOME/scikit_learn_data`.
    """
    if data_dir is None:
        data_dir = Path(get_sklearn_data_home()) / dataset_name
    else:
        data_dir = Path(data_dir) / dataset_name
        data_dir.mkdir(parents=True, exist_ok=True)

    return str(data_dir)


def hf_hub_download(data_home_dir: str, dataset_name: str, verbose: bool) -> str:
    """
    Download the given scikit-fingerprints dataset from HuggingFace Hub.
    Returns the absolute path to the directory with downloaded dataset.
    """
    pbar_was_disabled = are_progress_bars_disabled()
    try:
        if not verbose:
            disable_progress_bars()

        return snapshot_download(
            f"scikit-fingerprints/{dataset_name}",
            repo_type="dataset",
            local_dir=data_home_dir,
            cache_dir=data_home_dir,
        )
    finally:
        if not pbar_was_disabled:
            enable_progress_bars()


def get_mol_strings_and_labels(
    df: pd.DataFrame, mol_type: str = "SMILES"
) -> tuple[list[str], np.ndarray]:
    """
    Extract molecule strings (either SMILES or aminoacid sequences) and labels (one
    or more) from the given DataFrame.

    If ``mol_type`` is ``"SMILES"``, assumes that
    SMILES strings are in the "SMILES" column, another option is ``"aminoseq"``, which
    works similarly, but for aminoacid sequences. All other columns are taken as labels.

    If there is only a single task, labels are returned as a vector.
    """
    if mol_type == "SMILES":
        mol_strings = df.pop("SMILES").tolist()
    elif mol_type == "aminoseq":
        mol_strings = df.pop("aminoseq").tolist()
    else:
        raise ValueError(f"mol_type {mol_type} not recognized")

    # make sure we remove both columns if present
    df = df.drop(columns=["SMILES", "aminoseq"], errors="ignore")

    labels = df.values
    if labels.shape[1] == 1:
        labels = labels.ravel()

    return mol_strings, labels



"""Performance metrics useful in chemoinformatics."""

from .auroc import auroc_score
from .multioutput import (
    multioutput_accuracy_score,
    multioutput_auprc_score,
    multioutput_auroc_score,
    multioutput_balanced_accuracy_score,
    multioutput_cohen_kappa_score,
    multioutput_f1_score,
    multioutput_matthews_corr_coef,
    multioutput_mean_absolute_error,
    multioutput_mean_squared_error,
    multioutput_precision_score,
    multioutput_recall_score,
    multioutput_root_mean_squared_error,
    multioutput_spearman_correlation,
)
from .spearman import spearman_correlation
from .utils import extract_pos_proba
from .virtual_screening import bedroc_score, enrichment_factor, rie_score



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class MQNsFingerprint(BaseFingerprintTransformer):
    """
    Molecular Quantum Numbers (MQNs) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, which
    uses counts of 42 simple structural features [1]_.

    Features can be divided into four categories:

    - atom counts (12 features):

        - atoms: C, F, Cl, Br, I, S, P
        - acyclic and cyclic nitrogens
        - acyclic and cyclic oxygens
        - heavy atom count (HAC)

    - bond counts (7 features):

        - acyclic and cyclic bonds: single, double, triple
        - rotatable bonds

    - polarity counts (6 features)

        - H-bond acceptor sites and atoms
        - H-bond donor sites and atoms
        - negative and positive charges (predicted at pH 7)

    - topology counts (for H-depleted graph, 17 features):

        - acyclic nodes: single, di-, tri- and tetravalent
        - cyclic nodes: di, tri- and tetravalent
        - rings with size 3, 4, ..., 9, 10 or larger
        - nodes and bonds shared by 2 or more rings

    RDKit implementation is used, and may differ slightly from the original one in
    terms of polarity due to different donor and acceptor definitions.

    Note that by default this fingerprint returns count, not bit features, in order
    to follow the original paper.

    Parameters
    ----------
    count : bool, default=True
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 42
        Number of output features, size of fingerprints.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Kong T. Nguyen, Lorenz C. Blum, Ruud van Deursen, Jean-Louis Reymond
        "Classification of Organic Molecules by Molecular Quantum Numbers"
        ChemMedChem Volume 4, Issue 11 p. 1803-1805
        <https://chemistry-europe.onlinelibrary.wiley.com/doi/abs/10.1002/cmdc.200900317>`_

    Examples
    --------
    >>> from skfp.fingerprints import MQNsFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = MQNsFingerprint()
    >>> fp
    MQNsFingerprint()

    >>> fp.transform(smiles)  # doctest: +ELLIPSIS
    array([[0, 0, 0, ..., 0, 0, 0],
           [2, 0, 0, ..., 0, 0, 0],
           [1, 0, 0, ..., 0, 0, 0],
           [2, 0, 0, ..., 0, 0, 0]],
          dtype=uint32)
    """

    def __init__(
        self,
        count: bool = True,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=42,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute MQNs fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, 42)
            Array with fingerprints.
        """
        return super().transform(X, copy)

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import MQNs_

        X = ensure_mols(X)

        X = np.array([MQNs_(mol) for mol in X])
        if not self.count:
            X = X > 0

        dtype = np.uint32 if self.count else np.uint8
        return csr_array(X, dtype=dtype) if self.sparse else np.array(X, dtype=dtype)



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class VSAFingerprint(BaseFingerprintTransformer):
    """
    VSA (Van der Waals Surface Area) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, which
    calculates the atomic contributions to the approximate van der Waals surface
    area (VSA / ASA), based on given atomc properties [1]_.

    To calculate VSA, one gets the contribution of each atom in the molecule to
    a molecular property (e.g. SLogP) along with the contribution of each atom to
    the approximate molecular surface area (VSA), assign the atoms to bins based on
    the property contributions, and then sum up the VSA contributions for each atom
    in a bin.

    Features based on atomic contributions can be used:

    - SlogP, calculated water-octanol partition coefficient, measures lipophilicity [2]_
    - SMR, molar refractivity, measures polarizability [2]_
    - PEOE, Gasteiger partal charges, measure direct electrostatic interactions [3]_
    - EState, electrotopological states, encode information about both the topological
      environment of that atom and the electronic interactions [4]_

    Histogram bins are based on feature distributions in molecular datasets. RDKit uses
    slightly different bins from the original paper, see [5]_ [6]_ for exact values.
    See also [7]_ for methods of interpreting the values.

    Parameters
    ----------
    variant : {"all", "all_original", "SlogP", "SMR", "PEOE", "EState"}, default="all_original"
        Which features to use for calculating VSA bins. "all_original" uses three
        features from the original paper [1]_, i.e. SlogP, SMR and PEOE. "all" also
        adds EState.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features. Depends on variant: "SlogP" (12), "SMR" (10),
        "PEOE" (14), "EState" (11), "all_original" (36) or "all" (47).

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Paul Labute
        "A widely applicable set of descriptors"
        Journal of Molecular Graphics and Modelling, Volume 18, Issues 4–5, 2000, Pages 464-477
        <https://www.sciencedirect.com/science/article/pii/S1093326300000681>`_

    .. [2] `Scott A. Wildman and Gordon M. Crippen
        "Prediction of Physicochemical Parameters by Atomic Contributions"
        J. Chem. Inf. Comput. Sci. 1999, 39, 5, 868–873
        <https://pubs.acs.org/doi/10.1021/ci990307l>`_

    .. [3] `Johann Gasteiger and Mario Marsili
        "Iterative partial equalization of orbital electronegativity — a rapid access to atomic charges"
        Tetrahedron, Volume 36, Issue 22, 1980, Pages 3219-3228
        <https://www.sciencedirect.com/science/article/abs/pii/0040402080801682>`_

    .. [4] `Lowell H. Hall and Lemont B. Kier
        "Electrotopological State Indices for Atom Types: A Novel Combination of Electronic,
        Topological, and Valence State Information"
        J. Chem. Inf. Comput. Sci. 1995, 35, 6, 1039–1045
        <https://pubs.acs.org/doi/10.1021/ci00028a014>`_

    .. [5] `RDKit SlogP, SMR and PEOE bin values
        <https://github.com/rdkit/rdkit/blob/41f5c82ccf36a1b339dbf00440d823379decc2c3/rdkit/Chem/MolSurf.py>`_

    .. [6] `RDKit EState bin values
        <https://github.com/rdkit/rdkit/blob/41f5c82ccf36a1b339dbf00440d823379decc2c3/rdkit/Chem/EState/EState_VSA.py>`_

    .. [7] `Gregory Landrum
        "What are the VSA Descriptors?"
        <https://greglandrum.github.io/rdkit-blog/posts/2023-04-17-what-are-the-vsa-descriptors.html>`_

    Examples
    --------
    >>> from skfp.fingerprints import VSAFingerprint
    >>> smiles = ["CCO", "CCN"]
    >>> fp = VSAFingerprint()
    >>> fp
    VSAFingerprint()

    >>> fp.transform(smiles)  # doctest: +SKIP
    array([[ 0.        , 11.71340936,  0.        ,  ...,  0.        ,  0.        , 0.        ],
           [ 5.73366748,  6.54475641,  0.        ,  ...,  0.        ,  0.        , 0.        ]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "variant": [
            StrOptions({"all", "all_original", "SlogP", "SMR", "PEOE", "EState"})
        ],
    }

    def __init__(
        self,
        variant: str = "all_original",
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=self._get_n_features_out(variant),
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.variant = variant

    def _get_n_features_out(self, variant: str) -> int:
        n_features_out = {
            "SlogP": 12,
            "SMR": 10,
            "PEOE": 14,
            "EState": 11,
            "all_original": 36,  # SlogP + SMR + PEOE
            "all": 47,  # all Labute + Estate
        }
        try:
            return n_features_out[variant]
        except KeyError:
            raise ValueError(f'Variant "{variant} not recognized"')

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.EState.EState_VSA import EState_VSA_
        from rdkit.Chem.rdMolDescriptors import PEOE_VSA_, SMR_VSA_, SlogP_VSA_

        mols = ensure_mols(X)

        vsa_funcs = {
            "SlogP": lambda mols: np.array([SlogP_VSA_(mol) for mol in mols]),
            "SMR": lambda mols: np.array([SMR_VSA_(mol) for mol in mols]),
            "PEOE": lambda mols: np.array([PEOE_VSA_(mol) for mol in mols]),
            "EState": lambda mols: np.array([EState_VSA_(mol) for mol in mols]),
        }

        if self.variant in vsa_funcs:
            X = vsa_funcs[self.variant](mols)
        elif self.variant == "all_original":
            X_slogp = vsa_funcs["SlogP"](mols)
            X_smr = vsa_funcs["SMR"](mols)
            X_peoe = vsa_funcs["PEOE"](mols)
            X = np.column_stack((X_slogp, X_smr, X_peoe))
        else:  # "all"
            X_slogp = vsa_funcs["SlogP"](mols)
            X_smr = vsa_funcs["SMR"](mols)
            X_peoe = vsa_funcs["PEOE"](mols)
            X_estate = vsa_funcs["EState"](mols)
            X = np.column_stack((X_slogp, X_smr, X_peoe, X_estate))

        return csr_array(X) if self.sparse else X



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import validate_params

from skfp.datasets.utils import fetch_dataset, get_mol_strings_and_labels


@validate_params(
    {
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_bbbp(
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_frame: bool = False,
    verbose: bool = False,
) -> Union[pd.DataFrame, tuple[list[str]], np.ndarray]:
    """
    Load and return the BBBP (Blood-Brain Barrier Penetration) dataset.

    The task is to predict blood-brain barrier penetration (barrier permeability)
    of small drug-like molecules [1]_ [2]_.

    ==================   ==============
    Tasks                             1
    Task type            classification
    Total samples                  2039
    Recommended split          scaffold
    Recommended metric            AUROC
    ==================   ==============

    Parameters
    ----------
    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_frame : bool, default=False
        If True, returns the raw DataFrame with columns: "SMILES", "label". Otherwise,
        returns SMILES as list of strings, and labels as a NumPy array (1D integer binary
        vector).

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : pd.DataFrame or tuple(list[str], np.ndarray)
        Depending on the ``as_frame`` argument, one of:
        - Pandas DataFrame with columns: "SMILES", "label"
        - tuple of: list of strings (SMILES), NumPy array (labels)

    References
    ----------
    .. [1] `Ines Filipa Martins et al.
        "A Bayesian Approach to in Silico Blood-Brain Barrier Penetration Modeling"
        J. Chem. Inf. Model. 2012, 52, 6, 1686–1697
        <https://pubs.acs.org/doi/10.1021/ci300124c>`_

    .. [2] `Zhenqin Wu et al.
        "MoleculeNet: a benchmark for molecular machine learning"
        Chem. Sci., 2018,9, 513-530
        <https://pubs.rsc.org/en/content/articlelanding/2018/sc/c7sc02664a>`_
    """
    df = fetch_dataset(
        data_dir, dataset_name="MoleculeNet_BBBP", filename="bbbp.csv", verbose=verbose
    )
    return df if as_frame else get_mol_strings_and_labels(df)



from collections.abc import Sequence
from numbers import Integral
from typing import Any, Optional, Union

import numpy as np
from numpy.random import Generator, RandomState
from rdkit.Chem import Mol
from sklearn.utils._param_validation import Interval, RealNotInt, validate_params

from skfp.model_selection.splitters.scaffold_split import _create_scaffold_sets
from skfp.model_selection.splitters.utils import (
    ensure_nonempty_subset,
    split_additional_data,
    validate_train_test_split_sizes,
    validate_train_valid_test_split_sizes,
)
from skfp.utils import ensure_mols
from skfp.utils.functions import get_data_from_indices


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "use_csk": ["boolean"],
        "return_indices": ["boolean"],
        "random_state": ["random_state"],
    },
    prefer_skip_nested_validation=True,
)
def randomized_scaffold_train_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    test_size: Optional[float] = None,
    use_csk: bool = False,
    return_indices: bool = False,
    random_state: Optional[Union[int, RandomState, Generator]] = None,
):
    # flake8: noqa: E501
    """
    Split using randomized groups of Bemis-Murcko scaffolds.

    This split uses randomly partitioned groups of Bemis-Murcko molecular scaffolds [1]_
    for splitting. This is a nondeterministic variant of scaffold split, introduced in
    the MoleculeNet [2]_ paper. It aims to verify the model generalization to new scaffolds,
    as an approximation to the time split, while also allowing multiple train-test splits.

    By default, core structure scaffolds are used (following RDKit), which include atom
    types. Original Bemis-Murcko approach uses the cyclic skeleton (CSK) of a molecule,
    replacing all atoms by carbons. It is also known as CSK [3]_, and can be used with
    `use_csk` parameter.

    This approach is known to have certain limitations. In particular, molecules with
    no rings will not get a scaffold, resulting in them being grouped together regardless
    of their structure.

    This variant is nondeterministic, and the scaffolds are randomly shuffled before
    being assigned to subsets (test set is created fist). This approach is also known
    as "balanced scaffold split", and typically leads to more optimistic evaluation than
    regular, deterministic scaffold split [4]_.

    If ``train_size`` and ``test_size`` are integers, they must sum up to the ``data``
    length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: sequence
        Additional sequences to be split alongside the main data, e.g. labels.

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set
        to 1 - test_size. If test_size is also None, it will be set to 0.8.

    test_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set
        to 1 - train_size. If train_size is also None, it will be set to 0.2.

    use_csk: bool, default=False
        Whether to use the molecule cyclic skeleton (CSK), instead of the core
        structure scaffold.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    random_state: int or NumPy Random Generator instance, default=0
        Seed for random number generator or random state that would be used for
        shuffling the scaffolds.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-test subsets of provided arrays. First two are lists of SMILES
        strings or RDKit ``Mol`` objects, depending on the input type. If `return_indices`
        is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Bemis, G. W., & Murcko, M. A.
        "The properties of known drugs. 1. Molecular frameworks."
        Journal of Medicinal Chemistry, 39(15), 2887-2893.
        <https://www.researchgate.net/publication/14493474_The_Properties_of_Known_Drugs_1_Molecular_Frameworks>`_

    .. [2] `Z. Wu, B. Ramsundar, E. N. Feinberg, J. Gomes, C. Geniesse, A. S. Pappu, K. Leswing, V. Pande
        "MoleculeNet: A Benchmark for Molecular Machine Learning."
        Chemical Science, 9(2), 513-530.
        <https://www.researchgate.net/publication/314182452_MoleculeNet_A_Benchmark_for_Molecular_Machine_Learning>`_

    .. [3] `Bemis-Murcko scaffolds and their variants
        <https://github.com/rdkit/rdkit/discussions/6844>`_

    .. [4] `R. Sun, H. Dai, A. Wei Yu
        "Does GNN Pretraining Help Molecular Representation?"
        Advances in Neural Information Processing Systems 35 (NeurIPS 2022).
        <https://proceedings.neurips.cc/paper_files/paper/2022/hash/4ec360efb3f52643ac43fda570ec0118-Abstract-Conference.html>`_
    """
    # flake8: noqa: E501
    train_size, test_size = validate_train_test_split_sizes(
        train_size, test_size, len(data)
    )

    scaffold_sets = _create_scaffold_sets(data, use_csk)
    rng = (
        random_state
        if isinstance(random_state, RandomState)
        else np.random.default_rng(random_state)
    )
    rng.shuffle(scaffold_sets)

    train_idxs: list[int] = []
    test_idxs: list[int] = []

    for scaffold_set in scaffold_sets:
        if len(test_idxs) < test_size:
            test_idxs.extend(scaffold_set)
        else:
            train_idxs.extend(scaffold_set)

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, test_idxs
        )
        return train_subset, test_subset, *additional_data_split
    else:
        return train_subset, test_subset


@validate_params(
    {
        "data": ["array-like"],
        "additional_data": ["tuple"],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "valid_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(Integral, 1, None, closed="left"),
            None,
        ],
        "use_csk": ["boolean"],
        "return_indices": ["boolean"],
        "random_state": ["random_state"],
    },
    prefer_skip_nested_validation=True,
)
def randomized_scaffold_train_valid_test_split(
    data: Sequence[Union[str, Mol]],
    *additional_data: Sequence,
    train_size: Optional[float] = None,
    valid_size: Optional[float] = None,
    test_size: Optional[float] = None,
    use_csk: bool = False,
    return_indices: bool = False,
    random_state: Optional[Union[int, RandomState, Generator]] = None,
):
    """
    Split using randomized groups of Bemis-Murcko scaffolds.

    This split uses randomly partitioned groups of Bemis-Murcko molecular scaffolds [1]_
    for splitting. This is a nondeterministic variant of scaffold split, introduced in
    the MoleculeNet [2]_ paper. It aims to verify the model generalization to new scaffolds,
    as an approximation to the time split, while also allowing multiple train-test splits.

    By default, core structure scaffolds are used (following RDKit), which include atom
    types. Original Bemis-Murcko approach uses the cyclic skeleton (CSK) of a molecule,
    replacing all atoms by carbons. It is also known as CSK [3]_, and can be used with
    `use_csk` parameter.

    This approach is known to have certain limitations. In particular, molecules with
    no rings will not get a scaffold, resulting in them being grouped together regardless
    of their structure.

    This variant is nondeterministic, and the scaffolds are randomly shuffled before
    being assigned to subsets (in order: test, valid, train). This approach is also known
    as "balanced scaffold split", and typically leads to more optimistic evaluation than
    regular, deterministic scaffold split [4]_.

    If ``train_size``, ``valid_size`` and ``test_size`` are integers, they must sum up
    to the ``data`` length. If they are floating numbers, they must sum up to 1.

    Parameters
    ----------
    data : sequence
        A sequence representing either SMILES strings or RDKit ``Mol`` objects.

    additional_data: sequence
        Additional sequences to be split alongside the main data, e.g. labels.

    train_size : float, default=None
        The fraction of data to be used for the train subset. If None, it is set
        to 1 - test_size - valid_size. If valid_size is not provided, train_size
        is set to 1 - test_size. If train_size, test_size and valid_size aren't
        set, train_size is set to 0.8.

    valid_size : float, default=None
        The fraction of data to be used for the test subset. If None, it is set
        to 1 - train_size - valid_size. If train_size, test_size and valid_size
        aren't set, train_size is set to 0.1.

    test_size : float, default=None
        The fraction of data to be used for the validation subset. If None, it is
        set to 1 - train_size - valid_size. If valid_size is not provided, test_size
        is set to 1 - train_size. If train_size, test_size and valid_size aren't set,
        test_size is set to 0.1.

    use_csk: bool, default=False
        Whether to use the molecule cyclic skeleton (CSK), instead of the core
        structure scaffold.

    return_indices : bool, default=False
        Whether the method should return the input object subsets, i.e. SMILES strings
        or RDKit ``Mol`` objects, or only the indices of the subsets instead of the data.

    random_state: int or NumPy Random Generator instance, default=0
        Seed for random number generator or random state that would be used for
        shuffling the scaffolds.

    Returns
    ----------
    subsets : tuple[list, list, ...]
        Tuple with train-valid-test subsets of provided arrays. First three are lists of
        SMILES strings or RDKit ``Mol`` objects, depending on the input type. If
        `return_indices` is True, lists of indices are returned instead of actual data.

    References
    ----------
    .. [1] `Bemis, G. W., & Murcko, M. A.
        "The properties of known drugs. 1. Molecular frameworks."
        Journal of Medicinal Chemistry, 39(15), 2887-2893.
        <https://www.researchgate.net/publication/14493474_The_Properties_of_Known_Drugs_1_Molecular_Frameworks>`_

    .. [2] `Z. Wu, B. Ramsundar, E. N. Feinberg, J. Gomes, C. Geniesse, A. S. Pappu, K. Leswing, V. Pande
        "MoleculeNet: A Benchmark for Molecular Machine Learning."
        Chemical Science, 9(2), 513-530.
        <https://www.researchgate.net/publication/314182452_MoleculeNet_A_Benchmark_for_Molecular_Machine_Learning>`_

    .. [3] `Bemis-Murcko scaffolds and their variants
        <https://github.com/rdkit/rdkit/discussions/6844>`_

    .. [4] `R. Sun, H. Dai, A. Wei Yu
        "Does GNN Pretraining Help Molecular Representation?"
        Advances in Neural Information Processing Systems 35 (NeurIPS 2022).
        <https://proceedings.neurips.cc/paper_files/paper/2022/hash/4ec360efb3f52643ac43fda570ec0118-Abstract-Conference.html>`_
    """
    train_size, valid_size, test_size = validate_train_valid_test_split_sizes(
        train_size, valid_size, test_size, len(data)
    )

    scaffold_sets = _create_scaffold_sets(data, use_csk)
    rng = (
        random_state
        if isinstance(random_state, RandomState)
        else np.random.default_rng(random_state)
    )
    rng.shuffle(scaffold_sets)

    train_idxs: list[int] = []
    valid_idxs: list[int] = []
    test_idxs: list[int] = []

    for scaffold_set in scaffold_sets:
        if len(test_idxs) < test_size:
            test_idxs.extend(scaffold_set)
        elif len(valid_idxs) < valid_size:
            valid_idxs.extend(scaffold_set)
        else:
            train_idxs.extend(scaffold_set)

    ensure_nonempty_subset(train_idxs, "train")
    ensure_nonempty_subset(valid_idxs, "validation")
    ensure_nonempty_subset(test_idxs, "test")

    if return_indices:
        train_subset = train_idxs
        valid_subset = valid_idxs
        test_subset = test_idxs
    else:
        train_subset = get_data_from_indices(data, train_idxs)
        valid_subset = get_data_from_indices(data, valid_idxs)
        test_subset = get_data_from_indices(data, test_idxs)

    if additional_data:
        additional_data_split: list[Sequence[Any]] = split_additional_data(
            list(additional_data), train_idxs, valid_idxs, test_idxs
        )
        return train_subset, valid_subset, test_subset, *additional_data_split
    else:
        return train_subset, valid_subset, test_subset



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol, MolFromSmarts
from scipy.sparse import csr_array
from sklearn.utils._param_validation import InvalidParameterError

from skfp.bases.base_fp_transformer import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class BaseSubstructureFingerprint(BaseFingerprintTransformer):
    """
    Base class for substructure fingerprints.

    The implementation uses RDKit. Fingerprint tests for presence of provided
    molecular patterns. Number of features in fingerprint is equal to the number
    of patterns it was constructed from.

    This class is not meant to be used directly. If you want to use custom SMARTS
    patterns, inherit from this class and pass the ``patterns`` parameter to the
    parent constructor.

    Parameters
    ----------
    patterns : Sequence[str]
        Sequence of molecular patterns in SMARTS format.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of patterns. Equal to length of ``patterns``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "patterns": [list],
    }

    def __init__(
        self,
        patterns: Sequence[str],
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
        random_state: Optional[int] = 0,
    ):
        super().__init__(
            n_features_out=len(patterns),
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
            random_state=random_state,
        )
        self.patterns = self._compile_smarts_patterns(patterns)

    def _compile_smarts_patterns(self, patterns: Sequence[str]) -> list[Mol]:
        # we have to perform validation manually, since we want to compile SMARTS
        # patterns in constructor, and this happens before scikit-learn calls
        # parameter validation
        if not all(isinstance(pattern, str) for pattern in patterns):
            raise InvalidParameterError(
                "The 'patterns' parameter must be a sequence of SMARTS patterns."
            )
        if not patterns:
            raise InvalidParameterError(
                "The 'patterns' parameter must be a non-empty list of SMARTS patterns."
            )

        compiled_patterns = []
        for pattern in patterns:
            pattern_mol = MolFromSmarts(pattern)
            if not pattern_mol:
                raise InvalidParameterError(f"Got invalid SMARTS pattern: '{pattern}'")
            else:
                compiled_patterns.append(pattern_mol)

        return compiled_patterns

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute substructure fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects. If ``use_3D``
            is True, only Mol objects with computed conformations and with
            ``conf_id`` property are allowed.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.
        """
        return super().transform(X, copy)

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        X = ensure_mols(X)

        X = [
            [len(mol.GetSubstructMatches(pattern)) for pattern in self.patterns]
            for mol in X
        ]
        X = np.array(X, dtype=np.uint32)

        if not self.count:
            X = (X > 0).astype(np.uint8)

        return csr_array(X) if self.sparse else X



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids


class RDFFingerprint(BaseFingerprintTransformer):
    r"""
    RDF (Radial Distribution Function descriptors) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, where
    features are based on the radial distribution function (RDF) of interatomic
    distances.

    RDF function can be interpreted as the probability distribution of finding
    an atom in a spherical volume of given radius ``r``, and is defined for all
    atoms i and j with distance :math:`r_{ij}` in the molecule with N atoms as:

    .. math::

        RDF(r) = \sum_{i}^{N-1} \sum_{j > i}^N w_i * w_j * e^{B (r - r_{ij})^2}

    This results in a Gaussian distribution, centered around each distance
    :math:`r_{ij}`, with width depending on the smoothing parameter B, which is set
    to 100 (similar to DRAGON software). Radii between 1 and 30 (inclusive) are used,
    corresponding to distances from 1Å to 16Å.

    7 weighting variants are used, unweighted and 6 based on atomic features:
    unweighted, atomic mass, van der Waals volume, electronegativity, polarizability,
    ion polarity, and IState [1]_ [2]_. They are relative to the carbon, e.g. molecular
    weight is: MW(atom_type) / MW(carbon).

    This results in 210 features. They are invariant to translation and rotation,
    independent of molecule size, and unique for a given conformation. See [3]_ [4]_
    [5]_ [6]_ for details.

    Parameters
    ----------
    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 210
        Number of output features, size of fingerprints.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    See Also
    --------
    :class:`MORSEFingerprint` : Related fingerprint, which uses scattered electron
        intensity instead of radial distribution function.

    References
    ----------
    .. [1] RDKit IState discussion
        https://github.com/rdkit/rdkit/discussions/6122

    .. [2] RDKit IState implementation
        https://github.com/rdkit/rdkit/blob/df2b6c8a8c775748c1dcac83af0f92af434bab81/Code/GraphMol/Descriptors/MolData3Ddescriptors.cpp#L127

    .. [3] `DRAGON documentation
        "RDF descriptors"
        <http://chemgps.bmc.uu.se/help/dragonx/RDFdescriptors.html>`_

    .. [4] `Roberto Todeschini and Viviana Consonni
        "Molecular Descriptors for Chemoinformatics"
        <https://onlinelibrary.wiley.com/doi/book/10.1002/9783527628766>`_

    .. [5] `Guillaume Godin
        "3D	descriptors in RDKit"
        UGM 2017
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Godin_3D_Descriptors.pdf>`_

    .. [6] `Hemmer, Markus C., Valentin Steinhauer, and Johann Gasteiger
        "Deriving the 3D structure of organic molecules from their infrared spectra"
        Vibrational spectroscopy 19.1 (1999): 151-164
        <https://www.sciencedirect.com/science/article/pii/S0924203199000144>`_

    Examples
    --------
    >>> from skfp.fingerprints import RDFFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = RDFFingerprint()
    >>> fp
    RDFFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)  # doctest: +SKIP
    array([[1.930e+00, 2.070e-01, 0.000e+00, ..., 0.000e+00, 0.000e+00, 0.000e+00],
           [1.790e+00, 9.990e-01, 4.160e-01, ..., 0.000e+00, 0.000e+00, 0.000e+00],
           [1.150e-01, 0.000e+00, 0.000e+00, ..., 0.000e+00, 0.000e+00, 0.000e+00],
           [1.427e+00, 9.920e-01, 1.443e+00, ..., 0.000e+00, 0.000e+00, 0.000e+00]])
    """

    def __init__(
        self,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=210,
            requires_conformers=True,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import CalcRDF

        X = require_mols_with_conf_ids(X)
        X = [CalcRDF(mol, confId=mol.GetIntProp("conf_id")) for mol in X]
        return csr_array(X) if self.sparse else np.array(X)



from typing import Optional

from rdkit.Chem import FilterCatalog, Mol
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams

from skfp.bases.base_filter import BaseFilter


class NIHFilter(BaseFilter):
    """
    NIH filter.

    Designed to filter out molecules containing with undesirable functional groups,
    including reactive functionalities and medicinal chemistry exclusions.

    Rule definitions are available in the supplementary material of the original
    publications [1]_ [2]_ and in RDKit code [3]_.

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Ajit Jadhav et al.
        "Quantitative Analyses of Aggregation, Autofluorescence, and Reactivity
        Artifacts in a Screen for Inhibitors of a Thiol Protease"
        J. Med. Chem. 2010, 53, 1, 37–51
        <https://pubs.acs.org/doi/10.1021/jm901070c>`_

    .. [2] `Richard G. Doveston et al.
        "A unified lead-oriented synthesis of over fifty molecular scaffolds"
        Org. Biomol. Chem., 2015,13, 859-865
        <https://pubs.rsc.org/en/content/articlelanding/2015/ob/c4ob02287d>`_

    .. [3] `RDKit NIH filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/nih.in>`_

    Examples
    --------
    >>> from skfp.filters import NIHFilter
    >>> smiles = ["C", "C=P"]
    >>> filt = NIHFilter()
    >>> filt
    NIHFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self._filters = self._load_filters()

    def _load_filters(self) -> FilterCatalog:
        filter_rules = FilterCatalogParams.FilterCatalogs.NIH
        params = FilterCatalog.FilterCatalogParams()
        params.AddCatalog(filter_rules)
        filters = FilterCatalog.FilterCatalog(params)
        return filters

    def _apply_mol_filter(self, mol: Mol) -> bool:
        errors = len(self._filters.GetMatches(mol))
        return not errors or (self.allow_one_violation and errors == 1)



import hashlib
import itertools
import struct
from collections import defaultdict
from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from datasketch import MinHash
from rdkit.Chem import MolToSmiles, PathToSubmol
from rdkit.Chem.rdchem import Mol
from rdkit.Chem.rdmolops import FindAtomEnvironmentOfRadiusN, GetDistanceMatrix
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class MAPFingerprint(BaseFingerprintTransformer):
    """
    MinHashed Atom Pair fingerprint (MAP).

    Implementation is based on the official MAP4 paper and code [1]_ [2]_. This is a
    hashed fingerprint, using the ideas from Atom Pair and SECFP fingerprints.

    It computes fragments based on pairs of atoms, using circular
    substructures around each atom represented with SMILES (like SECFP) and length
    of shortest path between them (like Atom Pair), and then hashes the resulting
    triplet using MinHash fingerprint.

    Subgraphs are created around each atom with increasing radius, starting
    with just an atom itself. It is then transformed into a canonical SMILES.
    In each iteration, it is increased by another atom (one "hop" on the graph).

    For each pair of atoms, length of the shortest path is created. Then, for
    each radius from 1 up to the given maximal radius, the triplet is created:
    (atom 1 SMILES, shortest path length, atom 2 SMILES). They are then hashed
    using MinHash algorithm.

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors. Depending on the ``variant`` argument, those are either
        raw hashes, bits, or counts. Must be positive.

    radius : int, default=2
        Number of iterations performed, i.e. maximum radius of resulting subgraphs.
        Another common notation uses diameter, therefore MAP4 has radius 2.

    variant : {"raw_hashes", "bit", "count"}, default="bit"
        Which variant to fingerprint to use. ``"raw_hashes"`` returns MinHash values.
        ``"bit"`` folds values into binary vector, and ``"count"`` uses folding with
        counting.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`AtomPairFingerprint` : Related fingerprint, which uses atom invariants and
        folding.

    :class:`SECFPFingerprint` : Related fingerprint, which only uses SMILES of circular
        subgraph around each atom.

    References
    ----------
    .. [1] `Alice Capecchi, Daniel Probst and Jean-Louis Reymond
        "One molecular fingerprint to rule them all: drugs, biomolecules, and the metabolome"
        J Cheminform 12, 43 (2020)
        <https://jcheminf.biomedcentral.com/articles/10.1186/s13321-020-00445-4>`_

    .. [2] `https://github.com/reymond-group/map4`

    Examples
    --------
    >>> from skfp.fingerprints import MAPFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = MAPFingerprint()
    >>> fp
    MAPFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "radius": [Interval(Integral, 0, None, closed="left")],
        "variant": [StrOptions({"bit", "count", "raw_hashes"})],
    }

    def __init__(
        self,
        fp_size: int = 1024,
        radius: int = 2,
        variant: str = "bit",
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
        random_state: Optional[int] = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
            random_state=random_state,
        )
        self.fp_size = fp_size
        self.radius = radius
        self.variant = variant

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        X = ensure_mols(X)
        X = np.stack(
            [self._calculate_single_mol_fingerprint(mol) for mol in X], dtype=int
        )

        if self.variant == "bit":
            X = (X > 0).astype(np.uint8)
        elif self.variant == "count":
            X = X.astype(np.uint32)

        return csr_array(X) if self.sparse else np.array(X)

    def _calculate_single_mol_fingerprint(self, mol: Mol) -> np.ndarray:
        atoms_envs = self._get_atom_envs(mol)
        shingles = self._get_atom_pair_shingles(mol, atoms_envs)

        if self.variant == "raw_hashes":
            encoder = MinHash(num_perm=self.fp_size, seed=self.random_state)
            encoder.update_batch(shingles)
            fp = encoder.digest()
        else:
            # bit/count folded version from original MAP4 and MHFP implementation
            hashes = [self._get_hash(shingle) for shingle in shingles]
            bits = [hash_val % self.fp_size for hash_val in hashes]
            fp = np.bincount(bits, minlength=self.fp_size)

        return fp

    def _get_atom_envs(self, mol: Mol) -> dict[int, list[Optional[str]]]:
        """
        For each atom get its environment, i.e. radius-hop neighborhood.
        """
        atoms_env = defaultdict(list)
        for atom in mol.GetAtoms():
            idx = atom.GetIdx()
            atom_envs = [
                self._find_neighborhood(mol, idx, r) for r in range(1, self.radius + 1)
            ]
            atoms_env[idx].extend(atom_envs)

        return atoms_env

    def _find_neighborhood(
        self, mol: Mol, atom_idx: int, n_radius: int
    ) -> Optional[str]:
        """
        Get the radius-hop neighborhood for a given atom. If there is no neighborhood
        of a given radius, e.g. 2-hop neighborhood for [Li]F with just two atoms,
        returns None.
        """
        try:
            env = FindAtomEnvironmentOfRadiusN(mol, atom_idx, n_radius)
        except ValueError:
            # "bad atom index" error happens if radius is larger than possible
            return None

        atom_map: dict[int, int] = dict()

        submol = PathToSubmol(mol, env, atomMap=atom_map)

        if atom_idx in atom_map:
            return MolToSmiles(
                submol,
                rootedAtAtom=atom_map[atom_idx],
                canonical=True,
                isomericSmiles=False,
            )
        else:
            return None

    def _get_atom_pair_shingles(self, mol: Mol, atoms_envs: dict) -> list[bytes]:
        """
        Gets a list of atom molecular shingles - circular structures around atom pairs,
        written as SMILES, separated by the bond distance between the two atoms along the
        shortest path.
        """
        shingles = []
        distance_matrix = GetDistanceMatrix(mol)
        num_atoms = mol.GetNumAtoms()
        shingle_dict: dict[str, int] = defaultdict(int)

        # Iterate through all pairs of atoms and radius. Shingles are stored in format:
        # (radius i neighborhood of atom A) | (distance between atoms A and B) | (radius i neighborhood of atom B)
        #
        # If we want to count the shingles, we increment their value in shingle_dict
        # After nested for-loop, all shingles, with their respective counts, will be added to atom_pairs list.
        for idx_1, idx_2 in itertools.combinations(range(num_atoms), 2):
            # distance_matrix consists of floats as integers, so they need to be converted to integers first
            dist = str(int(distance_matrix[idx_1][idx_2]))
            env_a = atoms_envs[idx_1]
            env_b = atoms_envs[idx_2]

            for i in range(self.radius):
                env_a_radius = env_a[i]
                env_b_radius = env_b[i]

                # can be None if we couldn't get atom neighborhood of given radius
                if not env_a_radius or not env_b_radius:
                    continue

                ordered = sorted([env_a_radius, env_b_radius])
                shingle = f"{ordered[0]}|{dist}|{ordered[1]}"

                if self.variant == "count":
                    shingle_dict[shingle] += 1
                else:
                    shingles.append(shingle)

        if self.variant == "count":
            # shingle in format:
            # (radius i neighborhood of atom A) | (distance between atoms A and B) | \
            # (radius i neighborhood of atom B) | (shingle count)
            shingle_count = [
                f"{shingle}|{shingle_count}"
                for shingle, shingle_count in shingle_dict.items()
            ]
            shingles.extend(shingle_count)

        # convert strings to bytes for hashing
        shingles = [shingle.encode() for shingle in shingles]

        return shingles

    def _get_hash(self, shingle: bytes) -> int:
        hash_bytes = hashlib.sha1(shingle, usedforsecurity=False).digest()
        hash_value = struct.unpack("<I", hash_bytes[:4])[0]
        return hash_value



from typing import Union

from rdkit.Chem import Mol, rdMolDescriptors

from skfp.bases.base_filter import BaseFilter


class RuleOfVeberFilter(BaseFilter):
    """
    Rule of Veber.

    The rule is designed to identify molecules likely to exhibit good oral
    bioavailability, based on the number of rotatable bonds and the topological
    polar surface area (TPSA) [1]_.

    Molecule must fulfill conditions:

    - number of rotatable bonds <= 10
    - TPSA <= 140

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Veber, D.F, Johnson S. R., Cheng H., Smith B. R., Ward K. W. , Kopple K. D.
        "Molecular Properties That Influence the Oral Bioavailability of Drug Candidates."
        J Med Chem. 2002 Jun 6;45(12):2615-23.
        <https://pubmed.ncbi.nlm.nih.gov/12036371/>`_

    Examples
    ----------
    >>> from skfp.filters import RuleOfVeberFilter
    >>> smiles = ["CC=O", "CC(C)[C@@H](CC1=CC(=C(C=C1)OC)OCCCOC)C[C@@H]([C@H](C[C@@H](C(C)C)C(=O)NCC(C)(C)C(=O)N)O)N"]
    >>> filt = RuleOfVeberFilter()
    >>> filt
    RuleOfVeberFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CC=O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation, return_indicators, n_jobs, batch_size, verbose
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            rdMolDescriptors.CalcNumRotatableBonds(mol) <= 10,
            rdMolDescriptors.CalcTPSA(mol) <= 140,
        ]

        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class PatternFingerprint(BaseFingerprintTransformer):
    """
    Pattern fingerprint.

    This fingerprint is an RDKit original [1]_. This is a hashed fingerprint,
    where fragments are created from very generic SMARTS patterns, taking into
    consideration only atom and bond types (including bond aromaticity).

    For each pattern, its occurrences are detected and hashed, including
    atom and bond types involved. The fact that particular pattern matched
    the molecule at all is also stored by hashing the pattern ID and size.

    Patterns are (last four are too complex to describe easily):

    - [*] (single atom)
    - [*]~[*] (directly bonded atoms)
    - [*]~[*]~[*] (three consecutive atoms)
    - [R]~1~[R]~[R]~1 (3-atom ring)
    - [R]~1[R]~[R]~[R]~1 (4-atom ring)
    - [R]~1~[R]~[R]~[R]~[R]~1 (5-atom ring)
    - [R]~1~[R]~[R]~[R]~[R]~[R]~1 (6-atom ring)
    - [*]~[*](~[*])~[*] (atom with 3 neighbors)
    - [*]~[*]~[*](~[*])~[*]
    - [R](@[R])(@[R])~[R]~[R](@[R])(@[R])
    - [*]~[R](@[R])@[R](@[R])~[*]
    - [*]~[R](@[R])@[R]@[R](@[R])~[*]

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    tautomers : bool, default=False
        Whether to include tautomeric bonds as a separate bond type when checking
        pattern matches.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Gregory Landrum
        "Fingerprints in the RDKit"
        UGM 2012
        <https://www.rdkit.org/UGM/2012/Landrum_RDKit_UGM.Fingerprints.Final.pptx.pdf>`_

    Examples
    --------
    >>> from skfp.fingerprints import PatternFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = PatternFingerprint()
    >>> fp
    PatternFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "tautomers": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        tautomers: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.tautomers = tautomers

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdmolops import PatternFingerprint as RDKitPatternFingerprint

        X = ensure_mols(X)
        X = [
            RDKitPatternFingerprint(
                mol, fpSize=self.fp_size, tautomerFingerprints=self.tautomers
            )
            for mol in X
        ]

        if self.sparse:
            return csr_array(X, dtype=np.uint8)
        else:
            return np.array(X, dtype=np.uint8)



from typing import Union

import numpy as np
from scipy.stats import spearmanr
from sklearn.metrics._regression import _check_reg_targets
from sklearn.utils._param_validation import StrOptions, validate_params


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
        "alternative": [StrOptions({"two-sided", "less", "greater"})],
        "return_p_value": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def spearman_correlation(
    y_true: Union[np.ndarray, list[float]],
    y_pred: Union[np.ndarray, list[float]],
    *,
    alternative: str = "two-sided",
    return_p_value: bool = False,
    equal_values_result: float = 1.0,
) -> float:
    """
    Spearman correlation.

    Calculates Spearman's rank correlation coefficient (rho). It is a nonparametric
    measure of rank correlation. High value means that values of two variables change
    with monotonic relationship.

    For constant inputs, i.e. exactly the same ``y_true`` and ``y_pred``, 1.0 is returned.
    This differs from SciPy behavior, which returns NaN in that situation. This can be
    controlled with ``equal_values_result`` parameter.

    Mainly intended for use as quality metric, where higher correlation between model
    prediction and ground truth is better. Can also be used for general correlation
    testing, by using ``alternative`` and ``return_p_value`` parameters.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_pred : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Estimated target values.

    alternative : {"two-sided", "less", "greater"}, default="two-sided"
        Alternative hypothesis. By default, it checks if ground truth and estimated
        values differ in any direction.

    return_p_value : bool, default=False
        Whether to return p-value instead of correlation value.

    equal_values_result : float, default=1.0
        What value to return if ``y_true`` and ``y_pred`` are equal. This overrides the
        ``return_p_value`` parameter if necessary.

    Returns
    -------
    score : float
        Spearman correlation value, or p-value if ``return_p_value`` is True.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import spearman_correlation
    >>> y_true = np.array([1, 2, 3])
    >>> y_pred = np.array([3, 4, 5])
    >>> spearman_correlation(y_true, y_pred)
    1.0
    >>> y_true = np.array([1, 2, 3, 4])
    >>> y_pred = np.array([4, 3, 2, 1])
    >>> spearman_correlation(y_true, y_pred)
    -1.0
    """

    y_type, y_true, y_pred, multioutput = _check_reg_targets(
        y_true, y_pred, multioutput=None
    )

    if np.all(np.isclose(y_true, y_pred)):
        return equal_values_result

    result = spearmanr(y_true, y_pred, alternative=alternative)
    return result.pvalue if return_p_value else result.statistic
